
from contextlib import contextmanager
from copy import deepcopy
import yaml
from collections import abc
from dataclasses import asdict as dtc_asdict
import tempfile
import os.path
from typing import Any, Iterator, List, Sequence, Tuple
from prometheus_client import REGISTRY

from dlt.common import json
from dlt.common.runners import pool_runner as runner, TRunArgs, TRunMetrics
from dlt.common.configuration import BasicConfiguration, make_configuration
from dlt.common.file_storage import FileStorage
from dlt.common.logger import process_internal_exception
from dlt.common.schema import Schema, normalize_schema_name
from dlt.common.typing import DictStrAny, StrAny
from dlt.common.utils import uniq_id, is_interactive
from dlt.common.sources import DLT_METADATA_FIELD, TItem, with_table_name

from dlt.extractors.extractor_storage import ExtractorStorageBase
from dlt.loaders.client_base import SqlClientBase, SqlJobClientBase
from dlt.unpacker.configuration import configuration as unpacker_configuration
from dlt.loaders.configuration import configuration as loader_configuration
from dlt.unpacker import unpacker
from dlt.loaders import loader
from dlt.pipeline.exceptions import InvalidPipelineContextException, MissingDependencyException, NoPipelineException, PipelineStepFailed, CannotRestorePipelineException, SqlClientNotAvailable
from dlt.pipeline.typing import PipelineCredentials


class Pipeline:
    def __init__(self, pipeline_name: str, log_level: str = "INFO") -> None:
        self.pipeline_name = pipeline_name
        self.default_schema_name: str = None
        self.root_path: str = None
        self.root_storage: FileStorage = None
        self.credentials: PipelineCredentials = None
        self.extractor_storage: ExtractorStorageBase = None
        self.state: DictStrAny = {}

        # addresses of pipeline components to be verified before they are run
        self._unpacker_instance: int = None
        self._loader_instance: int = None

        # patch config and initialize pipeline
        C = make_configuration(BasicConfiguration, BasicConfiguration, initial_values={
            "NAME": pipeline_name,
            "LOG_LEVEL": log_level
        })
        runner.initialize_runner(C, TRunArgs(True, 0))

    def create_pipeline(self, credentials: PipelineCredentials, working_dir: str = None, schema: Schema = None) -> None:
        # initialize root storage
        if not working_dir:
            working_dir = tempfile.mkdtemp()
        self.root_storage = FileStorage(working_dir, makedirs=True)

        # check if directory contains restorable pipeline
        try:
            self._restore_state()
            # wipe out the old pipeline
            self.root_storage.delete_folder("", recursively=True)
            self.root_storage.create_folder("")
        except FileNotFoundError:
            pass

        self.root_path = self.root_storage.storage_path
        self.credentials = credentials
        self._load_modules()
        self.extractor_storage = ExtractorStorageBase("1.0.0", True, FileStorage(os.path.join(self.root_path, "extractor"), makedirs=True), unpacker.unpack_storage)
        # create new schema if no default supplied
        if schema is None:
            schema = Schema(normalize_schema_name(self.pipeline_name))
        # persist schema with the pipeline
        self.set_default_schema(schema)
        # initialize empty state
        with self._managed_state():
            self.state = {
                "default_schema_name": self.default_schema_name,
                "pipeline_name": self.pipeline_name,
                # TODO: must come from resolved configuration
                "loader_client_type": credentials.CLIENT_TYPE,
                # TODO: must take schema prefix from resolved configuration
                "loader_schema_prefix": credentials.schema_prefix
            }

    def restore_pipeline(self, credentials: PipelineCredentials, working_dir: str) -> None:
        try:
            # do not create extractor dir - it must exist
            self.root_storage = FileStorage(working_dir, makedirs=False)
            # restore state
            try:
                self._restore_state()
            except FileNotFoundError:
                raise CannotRestorePipelineException(f"Cannot find a valid pipeline in {working_dir}")
            restored_name = self.state["pipeline_name"]
            if self.pipeline_name != restored_name:
                raise CannotRestorePipelineException(f"Expected pipeline {self.pipeline_name}, found {restored_name} pipeline instead")
            self.default_schema_name = self.state["default_schema_name"]
            credentials.schema_prefix = self.state["loader_schema_prefix"]
            self.root_path = self.root_storage.storage_path
            self.credentials = credentials
            self._load_modules()
            # schema must exist
            try:
                self.get_default_schema()
            except (FileNotFoundError):
                raise CannotRestorePipelineException(f"Default schema with name {self.default_schema_name} not found")
            self.extractor_storage = ExtractorStorageBase("1.0.0", True, FileStorage(os.path.join(self.root_path, "extractor"), makedirs=False), unpacker.unpack_storage)
        except CannotRestorePipelineException:
            pass

    def extract(self, items: Iterator[TItem], schema_name: str = None, table_name: str = None) -> None:
        # check if iterator or iterable is supported
        # if isinstance(items, str) or isinstance(items, dict) or not
        # TODO: check if schema exists
        with self._managed_state():
            default_table_name = table_name or self.pipeline_name
            # TODO: this is not very effective - we consume iterator right away, better implementation needed where we stream iterator to files directly
            all_items: List[DictStrAny] = []
            for item in items:
                # dispatch items by type
                if callable(item):
                    item = item()
                if isinstance(item, dict):
                    all_items.append(item)
                elif isinstance(item, abc.Sequence):
                    all_items.extend(item)

            try:
                self._extract_iterator(default_table_name, all_items)
            except Exception:
                raise PipelineStepFailed("extract", self.last_run_exception, runner.LAST_RUN_METRICS)

    def unpack(self, workers: int = 1, max_events_in_chunk: int = 100000) -> None:
        if is_interactive() and workers > 1:
            raise NotImplementedError("Do not use workers in interactive mode ie. in notebook")
        self._verify_unpacker_instance()
        # set runtime parameters
        unpacker.CONFIG.MAX_PARALLELISM = workers
        unpacker.CONFIG.MAX_EVENTS_IN_CHUNK = max_events_in_chunk
        # switch to thread pool for single worker
        unpacker.CONFIG.POOL_TYPE = "thread" if workers == 1 else "process"
        runner.run_pool(unpacker.CONFIG, unpacker.unpack)
        if runner.LAST_RUN_METRICS.has_failed:
            raise PipelineStepFailed("unpack", self.last_run_exception, runner.LAST_RUN_METRICS)

    def load(self, max_parallel_loads: int = 20) -> None:
        self._verify_loader_instance()
        loader.CONFIG.MAX_PARALLELISM = loader.CONFIG.MAX_PARALLEL_LOADS = max_parallel_loads
        runner.run_pool(loader.CONFIG, loader.load)
        if runner.LAST_RUN_METRICS.has_failed:
            raise PipelineStepFailed("load", self.last_run_exception, runner.LAST_RUN_METRICS)

    def flush(self) -> None:
        self.unpack()
        self.load()

    @property
    def working_dir(self) -> str:
        return os.path.abspath(self.root_path)

    @property
    def last_run_exception(self) -> BaseException:
        return runner.LAST_RUN_EXCEPTION

    def list_extracted_loads(self) -> Sequence[str]:
        self._verify_loader_instance()
        return unpacker.unpack_storage.list_files_to_unpack_sorted()

    def list_unpacked_loads(self) -> Sequence[str]:
        self._verify_loader_instance()
        return loader.load_storage.list_loads()

    def list_completed_loads(self) -> Sequence[str]:
        self._verify_loader_instance()
        return loader.load_storage.list_completed_loads()

    def list_failed_jobs(self, load_id: str) -> Sequence[Tuple[str, str]]:
        self._verify_loader_instance()
        failed_jobs: List[Tuple[str, str]] = []
        for file in loader.load_storage.list_archived_failed_jobs(load_id):
            if not file.endswith(".exception"):
                try:
                    failed_message = loader.load_storage.storage.load(file + ".exception")
                except FileNotFoundError:
                    failed_message = None
                failed_jobs.append((file, failed_message))
        return failed_jobs

    def get_default_schema(self) -> Schema:
        self._verify_unpacker_instance()
        return unpacker.schema_storage.load_store_schema(self.default_schema_name)

    def set_default_schema(self, new_schema: Schema) -> None:
        if self.default_schema_name:
            # delete old schema
            unpacker.schema_storage.remove_store_schema(self.default_schema_name)
            self.default_schema_name = None
        # save new schema
        unpacker.schema_storage.save_store_schema(new_schema)
        self.default_schema_name = new_schema.schema_name
        with self._managed_state():
            self.state["default_schema_name"] = self.default_schema_name

    def add_schema(self, aux_schema: Schema) -> None:
        unpacker.schema_storage.save_store_schema(aux_schema)

    def get_schema(self, name: str) -> Schema:
        return unpacker.schema_storage.load_store_schema(name)

    def remove_schema(self, name: str) -> None:
        unpacker.schema_storage.remove_store_schema(name)

    def sync_schema(self) -> None:
        self._verify_loader_instance()
        schema = unpacker.schema_storage.load_store_schema(self.default_schema_name)
        with loader.make_client(schema) as client:
            client.initialize_storage()
            client.update_storage_schema()

    def sql_client(self) -> SqlClientBase[Any]:
        self._verify_loader_instance()
        schema = unpacker.schema_storage.load_store_schema(self.default_schema_name)
        with loader.make_client(schema) as c:
            if isinstance(c, SqlJobClientBase):
                return c.sql_client
            else:
                raise SqlClientNotAvailable(loader.CONFIG.CLIENT_TYPE)

    def _configure_unpack(self) -> None:
        # create unpacker config
        unpacker_initial = {
            "UNPACKING_VOLUME_PATH": os.path.join(self.root_path, "unpacking"),
            "SCHEMA_VOLUME_PATH": os.path.join(self.root_path, "schemas"),
            "WRITER_TYPE": loader.supported_writer(),
            "ADD_EVENT_JSON": False
        }
        unpacker_initial.update(self._configure_runner())
        C = unpacker_configuration(initial_values=unpacker_initial)
        unpacker.configure(C, REGISTRY)
        self._unpacker_instance = id(unpacker.CONFIG)

    def _configure_load(self) -> None:
        # use credentials to populate loader config, it includes also client type
        loader_initial = dtc_asdict(self.credentials)
        loader_initial.update(self._configure_runner())
        loader_initial["DELETE_COMPLETED_JOBS"] = True
        C = loader_configuration(initial_values=loader_initial)
        try:
            loader.configure(C, REGISTRY, is_storage_owner=True)
        except ImportError:
            raise MissingDependencyException(
                f"{self.credentials.CLIENT_TYPE} loader",
                [f"python-dlt[{self.credentials.CLIENT_TYPE}]"],
                "Dependencies for specific loaders are available as extras of python-dlt"
            )
        self._loader_instance = id(loader.CONFIG)

    # def _only_active(f: TFun) -> TFun:
    #     def _wrapper(self) -> Any

    def _verify_loader_instance(self) -> None:
        if self._loader_instance is None:
            raise NoPipelineException()
        if self._loader_instance != id(loader.CONFIG):
            # TODO: consider restoring pipeline from current work dir instead
            raise InvalidPipelineContextException()

    def _verify_unpacker_instance(self) -> None:
        if self._loader_instance is None:
            raise NoPipelineException()
        if self._unpacker_instance != id(unpacker.CONFIG):
            # TODO: consider restoring pipeline from current work dir instead
            raise InvalidPipelineContextException()

    def _configure_runner(self) -> StrAny:
        return {
            "NAME": self.pipeline_name,
            "EXIT_ON_EXCEPTION": True,
            "LOADING_VOLUME_PATH": os.path.join(self.root_path, "loading")
        }

    def _load_modules(self) -> None:
        # configure loader
        self._configure_load()
        # configure unpacker
        self._configure_unpack()

    def _extract_iterator(self, default_table_name: str, items: Sequence[DictStrAny]) -> None:
        try:
            for idx, i in enumerate(items):
                if not isinstance(i, dict):
                    # TODO: convert non dict types into dict
                    items[idx] = i = {"v": i}
                if DLT_METADATA_FIELD not in i or i.get(DLT_METADATA_FIELD, None) is None:
                    # set default table name
                    with_table_name(i, default_table_name)

            load_id = uniq_id()
            self.extractor_storage.save_json(f"{load_id}.json", items)
            self.extractor_storage.commit_events(
                self.default_schema_name,
                self.extractor_storage.storage._make_path(f"{load_id}.json"),
                default_table_name,
                len(items),
                load_id
            )

            runner.LAST_RUN_METRICS = TRunMetrics(was_idle=False, has_failed=False, pending_items=0)
        except Exception as ex:
            process_internal_exception("extracting iterator failed")
            runner.LAST_RUN_METRICS = TRunMetrics(was_idle=False, has_failed=True, pending_items=0)
            runner.LAST_RUN_EXCEPTION = ex
            raise

    @contextmanager
    def _managed_state(self) -> Iterator[None]:
        backup_state = deepcopy(self.state)
        try:
            yield
        except Exception:
            # restore old state
            self.state.clear()
            self.state.update(backup_state)
            raise
        else:
            # persist old state
            self.root_storage.save("state.json", json.dumps(self.state))

    def _restore_state(self) -> None:
        self.state.clear()
        restored_state: DictStrAny = json.loads(self.root_storage.load("state.json"))
        self.state.update(restored_state)

    @staticmethod
    def save_schema_to_file(file_name: str, schema: Schema, remove_defaults: bool = True) -> None:
        with open(file_name, "w", encoding="utf-8") as f:
            f.write(schema.as_yaml(remove_defaults=remove_defaults))

    @staticmethod
    def load_schema_from_file(file_name: str) -> Schema:
        with open(file_name, "r", encoding="utf-8") as f:
            schema_dict: DictStrAny = yaml.safe_load(f)
        return Schema.from_dict(schema_dict)



from .basic_configuration import BasicConfiguration  # noqa: F401
from .unpacking_volume_configuration import UnpackingVolumeConfiguration, ProductionUnpackingVolumeConfiguration  # noqa: F401
from .loading_volume_configuration import LoadingVolumeConfiguration, ProductionLoadingVolumeConfiguration  # noqa: F401
from .schema_volume_configuration import SchemaVolumeConfiguration, ProductionSchemaVolumeConfiguration  # noqa: F401
from .pool_runner_configuration import PoolRunnerConfiguration, TPoolType  # noqa: F401
from .gcp_client_configuration import GcpClientConfiguration, GcpClientProductionConfiguration  # noqa: F401
from .postgres_configuration import PostgresConfiguration, PostgresProductionConfiguration  # noqa: F401
from .utils import make_configuration, TConfigSecret, open_configuration_file  # noqa: F401

from .exceptions import (  # noqa: F401
    ConfigEntryMissingException, ConfigEnvValueCannotBeCoercedException, ConfigIntegrityException, ConfigFileNotFoundException)



import datetime  # noqa: 251
from typing import Any, Optional, Type

from dlt.common import pendulum
from dlt.common.schema.typing import TDataType


_NOW_TS: float = pendulum.now().timestamp()
_FLOAT_TS_RANGE = 31536000.0  # seconds in year


def is_timestamp(t: Type[Any], v: Any) -> Optional[TDataType]:
    # autodetect int and float withing 1 year range of NOW
    if t in [int, float]:
        if v >= _NOW_TS - _FLOAT_TS_RANGE and v <= _NOW_TS + _FLOAT_TS_RANGE:
            return "timestamp"
    return None


def is_iso_timestamp(t: Type[Any], v: Any) -> Optional[TDataType]:
    # only strings can be converted
    if t is not str:
        return None
    if not v:
        return None
    # strict autodetection of iso timestamps
    try:
        dt = pendulum.parse(v, strict=True, exact=True)
        if isinstance(dt, datetime.datetime):
            return "timestamp"
    except Exception:
        pass
    return None




from typing import List, Optional, Type

from dlt.common.typing import StrAny
from dlt.common.configuration.utils import TConfigSecret, make_configuration, _get_key_value
from dlt.common.configuration import PoolRunnerConfiguration, TPoolType, PostgresConfiguration, PostgresProductionConfiguration, GcpClientConfiguration, GcpClientProductionConfiguration

from . import __version__


class DBTRunnerConfiguration(PoolRunnerConfiguration):
    POOL_TYPE: TPoolType = "none"
    STOP_AFTER_RUNS: int = 1
    PACKAGE_VOLUME_PATH: str = "_storage/dbt_runner"
    PACKAGE_REPOSITORY_URL: str = "https://github.com/scale-vector/rasa_semantic_schema_customization.git"
    PACKAGE_REPOSITORY_BRANCH: Optional[str] = None
    PACKAGE_REPOSITORY_SSH_KEY: TConfigSecret = TConfigSecret("")  # the default is empty value which will disable custom SSH KEY
    PACKAGE_PROFILES_DIR: str = "."
    PACKAGE_PROFILE_PREFIX: str = "rasa_semantic_schema"
    PACKAGE_SOURCE_TESTS_SELECTOR: str = "tag:prerequisites"
    PACKAGE_ADDITIONAL_VARS: Optional[StrAny] = None
    PACKAGE_RUN_PARAMS: List[str] = ["--fail-fast"]
    AUTO_FULL_REFRESH_WHEN_OUT_OF_SYNC: bool = True

    SOURCE_SCHEMA_PREFIX: str = None
    DEST_SCHEMA_PREFIX: Optional[str] = None

    @classmethod
    def check_integrity(cls) -> None:
        if cls.PACKAGE_REPOSITORY_SSH_KEY and cls.PACKAGE_REPOSITORY_SSH_KEY[-1] != "\n":
            # must end with new line, otherwise won't be parsed by Crypto
            cls.PACKAGE_REPOSITORY_SSH_KEY = TConfigSecret(cls.PACKAGE_REPOSITORY_SSH_KEY + "\n")
        if cls.STOP_AFTER_RUNS != 1:
            # always stop after one run
            cls.STOP_AFTER_RUNS = 1


class DBTRunnerProductionConfiguration(DBTRunnerConfiguration):
    PACKAGE_VOLUME_PATH: str = "/var/local/app"  # this is actually not exposed as volume
    PACKAGE_REPOSITORY_URL: str = None


def gen_configuration_variant(initial_values: StrAny = None) -> Type[DBTRunnerConfiguration]:
    # derive concrete config depending on env vars present
    DBTRunnerConfigurationImpl: Type[DBTRunnerConfiguration]
    DBTRunnerProductionConfigurationImpl: Type[DBTRunnerProductionConfiguration]

    if _get_key_value("PG_SCHEMA_PREFIX", type(str)):
        source_schema_prefix = _get_key_value("PG_SCHEMA_PREFIX", type(str))
        class DBTRunnerConfigurationPostgress(PostgresConfiguration, DBTRunnerConfiguration):
            SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerConfigurationImpl = DBTRunnerConfigurationPostgress

        class DBTRunnerProductionConfigurationPostgress(DBTRunnerProductionConfiguration, PostgresProductionConfiguration, DBTRunnerConfigurationPostgress):
            pass
            # SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerProductionConfigurationImpl = DBTRunnerProductionConfigurationPostgress

    else:
        source_schema_prefix = _get_key_value("DATASET", type(str))
        class DBTRunnerConfigurationGcp(GcpClientConfiguration, DBTRunnerConfiguration):
            SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerConfigurationImpl = DBTRunnerConfigurationGcp

        class DBTRunnerProductionConfigurationGcp(DBTRunnerProductionConfiguration, GcpClientProductionConfiguration, DBTRunnerConfigurationGcp):
            pass
            # SOURCE_SCHEMA_PREFIX: str = source_schema_prefix
        DBTRunnerProductionConfigurationImpl = DBTRunnerProductionConfigurationGcp

    return make_configuration(DBTRunnerConfigurationImpl, DBTRunnerProductionConfigurationImpl, initial_values=initial_values)




from typing import Iterator, Tuple, Callable, TYPE_CHECKING

from dlt.common.typing import TEvent, StrAny
if TYPE_CHECKING:
    from dlt.common.schema import Schema

# type definitions for json normalization function

# iterator of form ((table_name, parent_table), dict) must be returned from normalization function
TUnpackedRowIterator = Iterator[Tuple[Tuple[str, str], StrAny]]

# normalization function signature
TNormalizeJSONFunc = Callable[["Schema", TEvent, str], TUnpackedRowIterator]



from typing import Any, Iterator

from dlt.common import json
from dlt.common.runners.venv import Venv
from dlt.common.runners.stdout import iter_stdout
from dlt.common.typing import DictStrAny


def get_source(venv: Venv, command: str, *script_args: Any) -> Iterator[DictStrAny]:
    # use pipe iterator and mapping function to get dict iterator from pipe
    yield from map(lambda s: json.loads(s), iter_stdout(venv, command, *script_args))  # type: ignore



import semver

from dlt.common.json import json_typed_dumps
from dlt.common.typing import Any
from dlt.common.utils import uniq_id
from dlt.common.schema import normalize_schema_name
from dlt.common.file_storage import FileStorage
from dlt.common.storages.versioned_storage import VersionedStorage
from dlt.common.storages.unpacker_storage import UnpackerStorage


class ExtractorStorageBase(VersionedStorage):
    def __init__(self, version: semver.VersionInfo, is_owner: bool, storage: FileStorage, unpacker_storage: UnpackerStorage) -> None:
        self.unpacker_storage = unpacker_storage
        super().__init__(version, is_owner, storage)

    def create_temp_folder(self) -> str:
        tf_name = uniq_id()
        self.storage.create_folder(tf_name)
        return tf_name

    def save_json(self, name: str, d: Any) -> None:
        # saves json using typed encoder
        self.storage.save(name, json_typed_dumps(d))

    def commit_events(self, schema_name: str, processed_file_path: str, dest_file_stem: str, no_processed_events: int, load_id: str, with_delete: bool = True) -> str:
        # schema name cannot contain underscores
        if schema_name != normalize_schema_name(schema_name):
            raise ValueError(schema_name)

        dest_name = UnpackerStorage.build_unpack_file_name(schema_name, dest_file_stem, no_processed_events, load_id)
        # if no events extracted from tracker, file is not saved
        if no_processed_events > 0:
            # moves file to possibly external storage and place in the dest folder atomically
            self.storage.copy_cross_storage_atomically(
                self.unpacker_storage.storage.storage_path, UnpackerStorage.UNPACKING_FOLDER, processed_file_path, dest_name)

        if with_delete:
            self.storage.delete(processed_file_path)

        return dest_name



import jsonlines
from typing import Iterator

from dlt.common import json
from dlt.common.typing import DictStrAny, StrOrBytesPath


def get_source(path: StrOrBytesPath) -> Iterator[DictStrAny]:
    with open(path, "r", encoding="utf-8") as f:
        yield from jsonlines.Reader(f, loads=json.loads)
